﻿using System;

using System.Runtime.CompilerServices;

using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;

namespace OpenDelegateBenchmark {

	public class Ops {
		public int a;

		public int Add(int b) => a + b;

		[MethodImpl(MethodImplOptions.NoInlining)]
		public int AddNoInline(int b) => a + b;

		public int Sub(int b) => a - b;

		public long Sub(long b) => a - b;

		public double Sub(double b) => a - b;

		public static int StaticAdd(int a, int b) => a + b;

		[MethodImpl(MethodImplOptions.NoInlining)]
		public static int StaticAddNoInline(int a, int b) => a + b;
	}

	public class Program {

		private Ops ops;

		private Func<int, int, int> staticDelegate;

		private Func<int, int> instanceDelegate;

		private Func<Ops, int, int> openInstanceDelegate;

		[GlobalSetup]
		public void GlobalSetup() {
			ops = new Ops { a = 10 };

			staticDelegate = Ops.StaticAdd;
			instanceDelegate = ops.Add;

			openInstanceDelegate = (Func<Ops, int, int>) Delegate.CreateDelegate(
				typeof(Func<Ops, int, int>),
				typeof(Ops).GetMethod("Add")
			);
		}

		[Benchmark]
		public int Static() => Ops.StaticAdd(10, 20);

		[Benchmark]
		public int StaticNoInline() => Ops.StaticAddNoInline(10, 20);

		[Benchmark]
		public int Instance() => ops.Add(20);

		[Benchmark]
		public int InstanceNoInline() => ops.AddNoInline(20);

		[Benchmark]
		public int StaticDelegate() => staticDelegate(10, 20);

		[Benchmark]
		public int InstanceDelegate() => instanceDelegate(20);

		[Benchmark]
		public int OpenInstanceDelegate() => openInstanceDelegate(ops, 20);

		[Benchmark]
		public Func<int, int, int> CreateStaticDelegate() => Ops.StaticAdd;

		[Benchmark]
		public Func<int, int> CreateInstanceDelegate() => ops.Add;

		[Benchmark]
		public Func<Ops, int, int> CreateOpenInstanceDelegate() => (Func<Ops, int, int>) Delegate.CreateDelegate(
			typeof(Func<Ops, int, int>),
			typeof(Ops).GetMethod("Add")
		);

		[Benchmark]
		public Func<Ops, int, int> CreateOpenInstanceDelegateOverloaded() => (Func<Ops, int, int>) Delegate.CreateDelegate(
			typeof(Func<Ops, int, int>),
			typeof(Ops).GetMethod("Sub", new[] { typeof(int) })
		);

		static void Main(string[] args) {
			BenchmarkRunner.Run<Program>();

			var prog = new Program();
			prog.GlobalSetup();
			Console.WriteLine(prog.OpenInstanceDelegate());
			Console.WriteLine(prog.CreateOpenInstanceDelegateOverloaded()(new Ops { a = 100 }, 20));
		}
	}
}
